#include <stdio.h>
#include <unistd.h>
#include <inttypes.h>
#include <string.h>
#include <assert.h>
#include <sched.h>
#include <linux/limits.h>
#include <stdlib.h>
#include <ctype.h>
#include "aftermath.h"
#include "kmp.h"

extern char *program_invocation_name;
struct am_trace kmpt_trace;
struct kmpt_global_context kmpt_gctx;

static const char* kmpt_state_names[KMPT_NUM_STATES] = {
	"barrier",
	"critical",
	"single",
	"taskwait",
	"master",
	"loop",
	"task",
	"sequential execution",
	"parallel region"
};

static inline int isnumber(const char* s)
{
	if(*s == '\0')
		return 0;

	while(*s) {
		if(!isdigit(*s))
			return 0;

		s++;
	}

	return 1;
}

static int kmpt_worker_context_init(struct kmpt_worker_context* wctx,
				    struct am_event_set* event_set)
{
	wctx->event_set = event_set;
	wctx->sequential_execution = 0;
	wctx->setup = 0;
	wctx->region_nesting = 0;
	kmpt_wsstack_init(&wctx->ws_stack, KMPT_MAX_WSSTACK_SIZE);

	if(am_state_stack_init(&wctx->state_stack, KMPT_MAX_STATE_NESTING))
		return 1;

	return 0;
}

static void kmpt_worker_context_destroy(struct kmpt_worker_context* wctx)
{
	am_state_stack_destroy(&wctx->state_stack);
	kmpt_wsstack_destroy(&wctx->ws_stack);
	kmpt_papi_worker_finish(wctx);
}

static int kmpt_global_context_init(struct kmpt_global_context* gctx,
				    struct am_trace* trace,
				    long num_cpus)
{
	long cpu;

	gctx->worker_ctx = (struct kmpt_worker_context*)
		malloc(sizeof(struct kmpt_worker_context) * num_cpus);

	if(!gctx->worker_ctx)
		goto out_err;

	memset(gctx->worker_ctx, 0, sizeof(struct kmpt_worker_context) * num_cpus);

	for(cpu = 0; cpu < num_cpus; cpu++) {
		if(kmpt_worker_context_init(&gctx->worker_ctx[cpu],
					    am_trace_get_event_set(trace, cpu)))
		{
			goto out_err_ctx;
		}
	}

	gctx->trace = trace;
	gctx->num_cpus = num_cpus;

	return 0;

out_err_ctx:
	for(long c = 0; c < cpu; c++)
		kmpt_worker_context_destroy(&gctx->worker_ctx[c]);

	free(gctx->worker_ctx);
out_err:
	return 1;
}

void kmpt_global_context_destroy(struct kmpt_global_context* gctx)
{
	struct kmpt_worker_context* wctx;

	for(long cpu = 0; cpu < gctx->num_cpus; cpu++) {
		wctx = &gctx->worker_ctx[cpu];

		/* Trace the sequential state for the worker executing the
		 * sequential part after the last parallel region. */
		if(wctx->sequential_execution) {
			if(am_state_stack_pop_trace(&wctx->state_stack, wctx->event_set)) {
				fprintf(stderr, "Could not pop state.\n");
				exit(1);
			}
		}

		kmpt_worker_context_destroy(&gctx->worker_ctx[cpu]);
	}

	kmpt_papi_shutdown(gctx);

	free(gctx->worker_ctx);
}

/* Executes the part of the worker initialization that must be called from the
 * thread associated to the worker (e.g., time stamp synchronization). */
void kmpt_setup_worker(struct kmpt_worker_context* wctx)
{
	struct kmpt_global_context* gctx = kmpt_get_global_context();

	if(wctx->setup)
		return;

	if(am_event_set_sync_timestamp_offset(wctx->event_set, gctx->trace))
		KMP_FATAL(FunctionError, "am_event_set_sync_timestamp_offset()");

	if(kmpt_papi_worker_init(wctx, gctx))
		KMP_FATAL(FunctionError, "kmpt_papi_worker_init()");

	wctx->setup = 1;
}

/* Initialization function for Aftermath tracing. */
int kmpt_init(void)
{
	long num_cpus;
	char* strace_buffer_size;
	size_t trace_buffer_size = KMPT_DEFAULT_TRACE_BUFFER_SIZE;
	char* sevent_set_buffer_size;
	size_t event_set_buffer_size = KMPT_DEFAULT_EVENT_SET_BUFFER_SIZE;
	struct kmpt_worker_context* this_worker;

	if((strace_buffer_size = getenv("AFTERMATH_OPENMP_TRACE_BUFFER_SIZE"))) {
		if(!isnumber(strace_buffer_size)) {
			fprintf(stderr, "%s is not a valid trace buffer size.",
				strace_buffer_size);
			goto out_err;
		}

		sscanf(strace_buffer_size, "%zu", &trace_buffer_size);
	}

	if(am_trace_init(&kmpt_trace, trace_buffer_size))
		goto out_err;

	if((num_cpus = sysconf(_SC_NPROCESSORS_ONLN)) == -1) {
		fprintf(stderr, "Could not determine number of processors");
		goto out_dest;
	}

	if((sevent_set_buffer_size = getenv("AFTERMATH_OPENMP_EVENT_SET_BUFFER_SIZE"))) {
		if(!isnumber(sevent_set_buffer_size)) {
			fprintf(stderr, "%s is not a valid event set buffer size.",
				sevent_set_buffer_size);
			goto out_dest;
		}

		sscanf(sevent_set_buffer_size, "%zu", &event_set_buffer_size);
	}

	for(long cpu = 0; cpu < num_cpus; cpu++) {
		if(am_trace_register_cpu(&kmpt_trace, cpu, event_set_buffer_size)) {
			fprintf(stderr, "Could not add CPU %ld.", cpu);
			goto out_dest;
		}
	}

	if(am_trace_set_reference_timestamp(&kmpt_trace, kmpt_get_cpu())) {
		fprintf(stderr, "Could not set timestamp reference.");
		goto out_dest;
	}

	for(size_t i = 0; i < KMPT_NUM_STATES; i++)
		if(am_trace_register_state(&kmpt_trace, i, kmpt_state_names[i]))
			goto out_dest;

	if(kmpt_global_context_init(&kmpt_gctx, &kmpt_trace, num_cpus)) {
		fprintf(stderr, "Could not initialize global context.");
		goto out_dest;
	}

	if(kmpt_papi_init(&kmpt_gctx)) {
		fprintf(stderr, "Could not initialize PAPI.");
		goto out_dest;
	}

	atexit(kmpt_exit);

	/* Mark this worker as the worker executing the code before the first
	 * parallel region. */
	this_worker = kmpt_get_worker_context();
	kmpt_setup_worker(this_worker);
	this_worker->sequential_execution = 1;
	kmpt_state_push(KMPT_STATE_SEQUENTIAL);

	return 0;

out_dest:
	am_trace_destroy(&kmpt_trace);
out_err:
	return 1;
}

/* Finalization function for Aftermath tracing. */
void kmpt_exit(void)
{
	const char* env_file_name = getenv("AFTERMATH_TRACE_FILE");
	char trace_file_name[PATH_MAX];

	if(!env_file_name) {
		strncpy(trace_file_name, program_invocation_name, PATH_MAX);
		sprintf(trace_file_name, "%s.ost", basename(program_invocation_name));
	} else {
		sprintf(trace_file_name, "%s", env_file_name);
	}

	kmpt_global_context_destroy(&kmpt_gctx);

	if(am_trace_dump(&kmpt_trace, trace_file_name))
		fprintf(stderr, "Could not dump trace file.");

	am_trace_destroy(&kmpt_trace);
}

/* Return the CPU identifier of the calling worker. */
am_cpu_t kmpt_get_cpu(void)
{
	return sched_getcpu();
}

/* Add an instance of a for loop whose data needs to be shared among
 * all workers of a team. Must be called with team context lock held
 * by the caller. */
struct kmpt_omp_for_instance*
kmpt_team_context_add_for_instance(struct kmpt_team_context* tctx)
{
	struct kmpt_omp_for_instance* next;

	if(tctx->num_for_instances > KMPT_MAX_OMP_FOR-1)
		return NULL;

	next = &tctx->for_instances[tctx->num_for_instances];

	memset(next, 0, sizeof(*next));
	tctx->num_for_instances++;

	return next;
}

/* Find the team's current instance of a shared for loop. Returns NULL
 * if no instance of the loop with the address is found. */
struct kmpt_omp_for_instance*
kmpt_team_context_find_for_instance(struct kmpt_team_context* tctx,
				    void* addr)
{
	size_t i = 0;

	for(i = 0; i < tctx->num_for_instances; i++)
		if(tctx->for_instances[i].addr == addr)
			return &tctx->for_instances[i];

	return NULL;
}

/* Return the worker's context. Must be called by the worker
 * itself. */
struct kmpt_worker_context* kmpt_get_worker_context(void)
{
	am_cpu_t cpu = kmpt_get_cpu();
	return &kmpt_gctx.worker_ctx[cpu];
}

/* Get the global context. Can be called from anywhere. */
struct kmpt_global_context* kmpt_get_global_context(void)
{
	return &kmpt_gctx;
}

/* Called whenever a worker enters a new team. This can be either a
 * new parallel team or a new serialized team. The function is *not*
 * called when a worker of a serialized team enters a parallel region
 * and the team is serialized again. The function is executed within
 * the context of the thread and after setup of the new team. */
void kmpt_worker_enter_team(int tid, uint64_t seq_team_id)
{
	struct kmpt_worker_context* wctx = kmpt_get_worker_context();

	/* If this is the worker executing the code before the first parallel
	 * region pop the initial sequential execution state. */
	if(wctx->sequential_execution) {
		wctx->sequential_execution = 0;

		/* Pop the sequential execution state */
		kmpt_state_pop();
	}

	if(KMP_MASTER_TID(tid)) {
		/* The master needs to push its current static for id base as
		 * it will have to be restored upon return to the previous
		 * team in order to stay in sync with the other workers. */

		if(kmpt_static_fori_id_stack_push(&wctx->static_fori_id_stack,
						  wctx->curr_static_fori_id))
		{
			fprintf(stderr, "Could not push current ID for static loops.");
			exit(1);
		}
	} else {
		/* Ordinary workers just need to initialize the static
		 * for id base stack */
		kmpt_static_fori_id_stack_init(&wctx->static_fori_id_stack);
	}

	wctx->curr_static_fori_id = 0;
	wctx->region_nesting++;

	kmpt_state_push(KMPT_STATE_PARALLEL_REGION);
}

/* Called whenever a worker leaves a team. This can be either a
 * parallel team or a serialized team that whose parent is a parallel
 * team. The function is *not* called for workers of a serialized team
 * whose parent has also been serialized. The function is executed
 * within the context of the thread and the team the worker is
 * leaving. */
void kmpt_worker_leave_team(int tid, uint64_t seq_team_id)
{
	struct kmpt_worker_context* wctx = kmpt_get_worker_context();

	/* Pop the parallel region execution state */
	kmpt_state_pop();
	wctx->region_nesting--;

	if(KMP_MASTER_TID(tid)) {
		/* Restore static for id base of the previous team */
		if(kmpt_static_fori_id_stack_top(&wctx->static_fori_id_stack,
						 &wctx->curr_static_fori_id))
		{
			fprintf(stderr, "Could not pop ID base for static loops.\n");
			exit(1);
		}

		kmpt_static_fori_id_stack_pop(&wctx->static_fori_id_stack);

		/* The master of the top-level (non-nested) parallel region
		 * continues with sequential execution. */
		if(wctx->region_nesting == 0) {
			wctx->sequential_execution = 1;
			kmpt_state_push(KMPT_STATE_SEQUENTIAL);
		}
	}
}

/* Initialize a team context, to be called on the creation of a
 * team */
void kmpt_team_context_init(struct kmpt_team_context* tctx)
{
	tctx->num_for_instances = 0;
	tctx->seq_team_id = KMP_TEST_THEN_INC32(&__kmpt_seq_team_id) + 1;
}

void kmpt_state_push(am_state_t state)
{
	struct kmpt_worker_context* wctx = kmpt_get_worker_context();

	if(am_state_stack_push_trace(&wctx->state_stack, wctx->event_set, state)) {
		fprintf(stderr, "Could not push state.");
		exit(1);
	}

	kmpt_papi_trace(wctx);
}

void kmpt_state_pop(void)
{
	struct kmpt_worker_context* wctx = kmpt_get_worker_context();

	if(am_state_stack_pop_trace(&wctx->state_stack, wctx->event_set)) {
		fprintf(stderr, "Could not pop state.\n");
		exit(1);
	}

	kmpt_papi_trace(wctx);
}

void kmpt_state_try_pop(void)
{
	struct kmpt_worker_context* wctx = kmpt_get_worker_context();
	int err = 0;

	am_state_stack_try_pop_trace(&wctx->state_stack, wctx->event_set, &err);

	if(err) {
		fprintf(stderr, "Could not pop state.\n");
		exit(1);
	}

	kmpt_papi_trace(wctx);
}
