#ifndef __AFTERMATH_H__
#define __AFTERMATH_H__

#include <aftermath/trace/trace.h>
#include <aftermath/trace/state_stack.h>
#include "aftermath_wsstack.h"
#include "aftermath_papi.h"

#define KMPT_DEFAULT_TRACE_BUFFER_SIZE (1 << 20)
#define KMPT_DEFAULT_EVENT_SET_BUFFER_SIZE (1 << 18)
#define KMPT_MAX_OMP_FOR_NESTING 32
#define KMPT_MAX_WSSTACK_SIZE (1 << 16)
#define KMPT_MAX_OMP_FOR 512
#define KMPT_MAX_STATE_NESTING 32

enum kmpt_state {
	KMPT_STATE_BARRIER = 0,
	KMPT_STATE_CRITICAL,
	KMPT_STATE_SINGLE,
	KMPT_STATE_TASKWAIT,
	KMPT_STATE_MASTER,
	KMPT_STATE_LOOP,
	KMPT_STATE_TASK,
	KMPT_STATE_SEQUENTIAL,
	KMPT_STATE_PARALLEL_REGION,
	KMPT_NUM_STATES
};

/* A static for instance id stack can be used by a worker to save and
 * restore integer values used for the generation of instance
 * ids. This allows the worker to keep locally generated for instance
 * ids synchronized with the team mates of all teams it is
 * involved. */
struct kmpt_static_fori_id_stack {
	uint64_t entries[KMPT_MAX_OMP_FOR_NESTING];
	size_t nesting;
};

/* Initialize a static for instance id stack. */
static inline void kmpt_static_fori_id_stack_init(struct kmpt_static_fori_id_stack* s)
{
	s->nesting = 0;
}

/* Push a new id base onto the stack. Returns 0 on success, 1
 * otherwise. */
static inline int kmpt_static_fori_id_stack_push(struct kmpt_static_fori_id_stack* s,
				     uint64_t id_base)
{
	if(s->nesting > KMPT_MAX_OMP_FOR_NESTING-1)
		return 1;

	s->entries[s->nesting] = id_base;
	s->nesting++;

	return 0;
}

/* Remove an id base from the stack */
static inline void kmpt_static_fori_id_stack_pop(struct kmpt_static_fori_id_stack* s)
{
	s->nesting--;
}

/* Get the topmost element of the stack. Returns 0 on success, 1
 * otherwise. */
static inline int
kmpt_static_fori_id_stack_top(struct kmpt_static_fori_id_stack* s, uint64_t* id_base)
{
	if(!s->nesting)
		return 1;

	*id_base = s->entries[s->nesting-1];
	return 0;
}

/* Generate a unique for instance ID for a loop with a dynamic
 * schedule composed by a timestamp and the cpu identifier:
 *
 * +-+---------------------------------------------------------+---------------+
 * |1| timestamp (47 bits)                                     | cpu (16 bits) |
 * +-+---------------------------------------------------------+---------------+
 *
 * The most significant bit is always 1 to avoid clashes with IDs
 * generated for loops with a static schedule.
 */
static inline am_omp_for_instance_id_t
kmpt_gen_dynamic_for_instance_id(am_timestamp_t t, am_cpu_t cpu)
{
	return ((t << 16) | (cpu & 0xFFFF)) | 0x8000000000000000;
}

/* Generate a unique for instance ID for a loop with a static schedule
 * composed by a sequential team id and the current base id for static
 * for loop instances of the team:
 *
 * +-+---------------------------------+---------------------------------------+
 * |0| sequential team id (31 bits)    | static for instance id base (32 bits) |
 * +-+---------------------------------+---------------------------------------+
 *
 * The most significant bit is always 0 to avoid clashes with IDs
 * generated for loops with a dynamic schedule.
 */
static inline am_omp_for_instance_id_t
kmpt_gen_static_for_instance_id(uint64_t seq_team_id, uint64_t id_base)
{
	return ((seq_team_id << 32) | (id_base & 0xFFFFFFFF))
		& 0x7FFFFFFFFFFFFFFF;
}

/* Generate a unique chunk set id composed by the current timestamp
 * and the cpu identifier:
 *
 * +---------------------------------------------------------------------------+
 * | timestamp (48 bits)                                       | cpu (16 bits) |
 * +---------------------------------------------------------------------------+
 */
static inline am_omp_for_chunk_set_id_t
kmpt_gen_chunk_set_id(am_timestamp_t t, am_cpu_t cpu)
{
	return (t << 16) | (cpu & 0xFFFF);
}

/* In-memory representation of an instance of a for loop. Mostly
 * identical with the on-disk representation. */
struct kmpt_omp_for_instance {
	am_omp_for_increment_t increment;
	am_omp_for_flags_t flags;
	am_omp_for_iterator_t lower_bound;
	am_omp_for_iterator_t upper_bound;
	void* addr;
	am_omp_worker_t num_workers;
	int is_executing;

	union {
		uint64_t id;

		struct {
			uint32_t id_num;
			uint32_t id_addr;
		};
	};
};

/* In-memory representation of an instance of chunk set part. Mostly
 * identical with the on-disk representation. */
struct kmpt_omp_for_chunk_set_part {
	am_omp_for_chunk_set_id_t chunk_set_id;
	am_timestamp_t start;
	am_timestamp_t end;
};

struct kmpt_team_context {
	struct kmpt_omp_for_instance for_instances[KMPT_MAX_OMP_FOR];
	size_t num_for_instances;
	uint64_t seq_team_id;
};

void kmpt_team_context_init(struct kmpt_team_context* tctx);

/* Reinitialize a team context, to be called each time a team is
 * reused. */
static void kmpt_team_context_reinit(struct kmpt_team_context* tctx)
{
	kmpt_team_context_init(tctx);
}

static inline am_omp_task_instance_id_t
kmpt_gen_task_instance_id(am_timestamp_t t, am_cpu_t cpu)
{
	return (t << 16) | (cpu & 0xFFFF);
}

struct kmpt_omp_task_instance {
	void* addr;
	union {
		uint64_t id;

		struct {
			uint32_t id_num;
			uint32_t id_addr;
		};
	};
};

struct kmpt_omp_task_instance_part {
	am_omp_task_instance_id_t task_instance_id;
	am_timestamp_t start;
	am_timestamp_t end;
};

/* Data structure for per-worker data needed for Aftermath tracing */
struct kmpt_worker_context {
	/* Stack of states */
	struct am_state_stack state_stack;

	/* Current chunk set part */
	struct kmpt_omp_for_chunk_set_part active_csp;

	/* Current task instance part */
	struct kmpt_omp_task_instance_part active_tip;

	/* Stack of worksharing construct */
	struct kmpt_wsstack ws_stack;

	/* Associated aftermath event set */
	struct am_event_set* event_set;

	/* Stack of integer bases for the generation of IDs of loops
	 * with a static schedule */
	struct kmpt_static_fori_id_stack static_fori_id_stack;

	/* Current integer base for static for loop ids */
	uint64_t curr_static_fori_id;

	/* Boolean preventing the tracing of empty static loops executions */
	int empty_loop;

	/* Boolean, indicating whether the associated worker has been set up
	 * with a call to kmpt_setup_this_worker() from its thread. */
	int setup;

	/* Indicates the nesting for parallel regions for this worker. */
	int region_nesting;

	/* Boolean indicating that this worker is currently the executing only
	 * one executing the program (e.g., at the very beginning of the
	 * execution before the first parallel region or between parallel
	 * regions). */
	int sequential_execution;

	/* PAPI fields */
	KMPT_PAPI_WORKER_FIELDS;
};

/* Root data structure for Aftermath tracing in the run-time. */
struct kmpt_global_context {
	/* Array per cpu */
	struct kmpt_worker_context* worker_ctx;

	/* Trace file */
	struct am_trace* trace;

	/* Total number of CPUs involved in execution */
	long num_cpus;

	/* PAPI fields */
	KMPT_PAPI_GLOBAL_FIELDS;
};

/* Restore the for instance at the top of the stack. Returns 0 on success,
 * otherwise 1.
 */
static inline int kmpt_wsstack_restore_top_for(struct kmpt_worker_context* wctx,
					       am_timestamp_t now)
{
	struct kmpt_wsstack* s = &wctx->ws_stack;
	struct kmpt_wsstack_for_entry* top;

	if(!(top = (struct kmpt_wsstack_for_entry*)kmpt_wsstack_top(s)))
		return 1;

	wctx->active_csp.chunk_set_id = top->cs_id;
	wctx->active_csp.start = now;

	return 0;
}

/* Restore the task instance at the top of the stack. Returns 0 on success,
 * otherwise 1.
 */
static inline int kmpt_wsstack_restore_top_task(struct kmpt_worker_context* wctx,
						am_timestamp_t now)
{
	struct kmpt_wsstack* s = &wctx->ws_stack;
	struct kmpt_wsstack_task_entry* top;

	if(!(top = (struct kmpt_wsstack_task_entry*)kmpt_wsstack_top(s)))
		return 1;

	wctx->active_tip.task_instance_id = top->tiid;
	wctx->active_tip.start = now;

	return 0;
}

/* Restore the worksharing construct at the top of the stack. Returns 0 on
 * success, otherwise 1.
 */
static inline int
kmpt_wsstack_restore_top(struct kmpt_worker_context* wctx, am_timestamp_t now)
{
	struct kmpt_wsstack* s = &wctx->ws_stack;

	switch(kmpt_wsstack_top_type(s)) {
		case KMPT_WSSTACK_ENTRY_TYPE_FOR:
			return kmpt_wsstack_restore_top_for(wctx, now);

		case KMPT_WSSTACK_ENTRY_TYPE_TASK:
			return kmpt_wsstack_restore_top_task(wctx, now);

		case KMPT_WSSTACK_ENTRY_TYPE_MAX:
		case KMPT_WSSTACK_EMPTY:
			return 1;
	}

	/* Cannot happen */
	return 0;
}

/* Terminate and trace the current for chunk set part */
static inline int
kmpt_wsstack_terminate_top_for(struct kmpt_worker_context* wctx,
			       am_timestamp_t now)
{
	struct kmpt_wsstack* s = &wctx->ws_stack;
	struct kmpt_wsstack_for_entry* top;

	if(!(top = kmpt_wsstack_top_for(s)))
		return 1;

	wctx->active_csp.end = now;

	if(am_event_set_trace_omp_for_chunk_set_part(wctx->event_set,
						     wctx->active_csp.chunk_set_id,
						     wctx->active_csp.start,
						     wctx->active_csp.end))
	{
		fprintf(stderr, "Could not trace for chunk set part.\n");
		return 1;
	}

	return 0;
}

/* Terminate and trace the current task instance part */
static inline int
kmpt_wsstack_terminate_top_task(struct kmpt_worker_context* wctx,
				am_timestamp_t now)
{
	struct kmpt_wsstack* s = &wctx->ws_stack;
	struct kmpt_wsstack_task_entry* top;

	if(!(top = kmpt_wsstack_top_task(s)))
		return 1;

	wctx->active_tip.end = now;

	if(am_event_set_trace_omp_task_instance_part(wctx->event_set,
						     wctx->active_tip.task_instance_id,
						     wctx->active_tip.start,
						     wctx->active_tip.end))
	{
		fprintf(stderr, "Could not trace task instance part.\n");
		return 1;
	}

	return 0;
}

/* Terminate and trace the current worksharing entry. */
static inline int
kmpt_wsstack_terminate_top(struct kmpt_worker_context* wctx, am_timestamp_t now)
{
	struct kmpt_wsstack* s = &wctx->ws_stack;

	switch(kmpt_wsstack_top_type(s)) {
		case KMPT_WSSTACK_EMPTY:
			return 0;

		case KMPT_WSSTACK_ENTRY_TYPE_FOR:
			return kmpt_wsstack_terminate_top_for(wctx, now);

		case KMPT_WSSTACK_ENTRY_TYPE_TASK:
			return kmpt_wsstack_terminate_top_task(wctx, now);

		default:
			return 1;
	}

	return 0;
}

int kmpt_init(void)__attribute__((constructor));
void kmpt_exit(void);
am_cpu_t kmpt_get_cpu(void);
void kmpt_setup_worker(struct kmpt_worker_context* wctx);

void kmpt_worker_enter_team(int tid, uint64_t seq_team_id);
void kmpt_worker_leave_team(int tid, uint64_t seq_team_id);

struct kmpt_worker_context* kmpt_get_worker_context(void);
struct kmpt_global_context* kmpt_get_global_context(void);

struct kmpt_omp_for_instance*
kmpt_team_context_add_for_instance(struct kmpt_team_context* gctx);

struct kmpt_omp_for_instance*
kmpt_team_context_find_for_instance(struct kmpt_team_context* gctx,
					    void* addr);

void kmpt_state_push(am_state_t state);
void kmpt_state_pop(void);
void kmpt_state_try_pop(void);

#endif
