#include "aftermath.h"
#include "aftermath_papi.h"
#include "kmp.h"
#include <string.h>
#include <stdio.h>
#include <pthread.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <kmp_i18n.h>

#ifdef KMPT_PAPI_ENABLED

static void kmpt_nop_microtask(int32_t * global_tid, int32_t * bound_tid, ...)
{
}

static unsigned long int kmpt_papi_get_cpu(void)
{
	return kmpt_get_cpu();
}

/* Parser for KMPT_PAPI_EVENTS. Sets kmpt_counter_names and kmpt_papi_num_events
 * accordingly. */
static void kmpt_papi_parse_env(struct kmpt_global_context* gctx)
{
	char* env;
	char* env_tmp;
	char* token;
	char* save_ptr;
	int i;

	if(!(env = getenv("KMPT_PAPI_COUNTERS"))) {
		gctx->papi_num_counters = 0;
		return;
	}

	env_tmp = strdup(env);

	for(token = strtok_r(env_tmp, ", ", &save_ptr), i = 0;
	    token != NULL;
	    token = strtok_r(NULL, ", ", &save_ptr), i++)
	{
		gctx->papi_counter_names[i] = strdup(token);
	}

	gctx->papi_num_counters = i;

	free(env_tmp);
}

/* Initialization of the PAPI library and construction of the event sets */
int kmpt_papi_init(struct kmpt_global_context* gctx)
{
	int ret;

	kmpt_papi_parse_env(gctx);

	if(gctx->papi_num_counters == 0)
		return 0;

	ret = PAPI_library_init(PAPI_VER_CURRENT);

	if (ret != PAPI_VER_CURRENT && ret > 0) {
		fprintf(stderr,"PAPI library version mismatch!\n");
		goto out_err;
	}

	if (ret < 0) {
		fprintf(stderr,"Could not init PAPI library: %s!\n",
			PAPI_strerror(ret));
		goto out_err;
	}

	if (PAPI_is_initialized() != PAPI_LOW_LEVEL_INITED) {
		fprintf(stderr, "Could not init PAPI library (low-level)!\n");
		goto out_err;
	}

	if ((ret = PAPI_thread_init(kmpt_papi_get_cpu)) != PAPI_OK) {
		fprintf(stderr, "PAPI: Could not init threads: %s\n",
			PAPI_strerror(ret));
		goto err_shutdown;
	}

	for(size_t i = 0; i < gctx->papi_num_counters; i++) {
		if(am_trace_register_counter(gctx->trace, i+PAPI_COUNTER_BASE,
					     gctx->papi_counter_names[i]))
		{
			fprintf(stderr,
				"PAPI: Could not register counter %s: %s\n",
				PAPI_strerror(ret));

			goto err_shutdown;
		}
	}

	/* Force initialization of the run-time before the first parallel
	 * region */
	__kmpc_fork_call(NULL, 1, kmpt_nop_microtask);

	return 0;

err_shutdown:
	kmpt_papi_shutdown(gctx);
out_err:
	return 1;
}

void kmpt_papi_shutdown(struct kmpt_global_context* gctx)
{
	if(gctx->papi_num_counters == 0)
		return;

	for(int i = 0; i < gctx->papi_num_counters; i++)
		free(gctx->papi_counter_names[i]);

	PAPI_shutdown();
}

int kmpt_papi_worker_init(struct kmpt_worker_context* wctx,
			  struct kmpt_global_context* gctx)
{
	int err, i;
	int component_idx = 0;

	/* No counters to trace? Nothing to do! */
	if(gctx->papi_num_counters == 0)
		return 0;

	wctx->papi_counters = (long long*)calloc(gctx->papi_num_counters,
						 sizeof(long long));

	if(!wctx->papi_counters)
		goto out_err;

	wctx->papi_event_mapping = (long long*)malloc(gctx->papi_num_counters *
						      sizeof(long long));

	if(!wctx->papi_event_mapping)
		goto out_err_ctrs;

	wctx->papi_count = 0;

	if(PAPI_register_thread() != PAPI_OK)
		goto out_err_mapping;

	wctx->papi_event_set = PAPI_NULL;

	/* Create event set */
	if ((err = PAPI_create_eventset(&wctx->papi_event_set)) != PAPI_OK) {
		fprintf(stderr, "Could not create event set for thread %d: %s!\n",
			kmpt_get_cpu(), PAPI_strerror(err));

		goto out_err_mapping;
	}

	component_idx = 0;

	/* Assign CPU component */
	if ((err = PAPI_assign_eventset_component(wctx->papi_event_set,
						  component_idx)) != PAPI_OK)
	{
		fprintf(stderr, "Could not assign event set to component 0 "
			"for thread %d: %s\n",
			kmpt_get_cpu(), PAPI_strerror(err));

		goto out_err_evs;
	}

	wctx->papi_num_counters = gctx->papi_num_counters;

	for(i = 0; i < gctx->papi_num_counters; i++) {
		wctx->papi_event_mapping[i] = i;

		/* Add event to event set */
		err = PAPI_add_named_event(wctx->papi_event_set,
					   gctx->papi_counter_names[i]);

		if(err != PAPI_OK) {
			fprintf(stderr, "Could not add event \"%s\" "
				"to event set: %s!\n",
				gctx->papi_counter_names[i],
				PAPI_strerror(err));
			goto out_err_evs;
		}
	}

	if(wctx->papi_num_counters > 0) {
		/* Reset */
		if ((err = PAPI_reset(wctx->papi_event_set)) != PAPI_OK) {
			fprintf(stderr, "Could not reset counters: %s!\n",
				PAPI_strerror(err));
			goto out_err_evs;
		}

		/* Start counting */
		if ((err = PAPI_start(wctx->papi_event_set)) != PAPI_OK) {
			fprintf(stderr, "Could not start counters: %s!\n",
				PAPI_strerror(err));
			goto out_err_evs;
		}

		wctx->papi_count = 1;
	}

	return 0;

out_err_evs:
	PAPI_cleanup_eventset(wctx->papi_event_set);
	PAPI_destroy_eventset(&wctx->papi_event_set);
out_err_mapping:
	free(wctx->papi_event_mapping);
	wctx->papi_event_mapping = NULL;
out_err_ctrs:
	free(wctx->papi_counters);
	wctx->papi_counters = NULL;
out_err:
	return 1;
}

void kmpt_papi_worker_finish(struct kmpt_worker_context* wctx)
{
	if(wctx->papi_num_counters == 0)
		return;

	PAPI_stop(wctx->papi_event_set, wctx->papi_counters);
	PAPI_cleanup_eventset(wctx->papi_event_set);
	PAPI_destroy_eventset(&wctx->papi_event_set);
	free(wctx->papi_event_mapping);
	free(wctx->papi_counters);
}

int kmpt_papi_trace_timestamp(struct kmpt_worker_context* wctx,
			      am_timestamp_t timestamp)
{
	if(wctx->papi_num_counters == 0 || !wctx->papi_count)
		return 0;

	if(PAPI_accum(wctx->papi_event_set, wctx->papi_counters) != PAPI_OK) {
		fprintf(stderr, "Could not read counters for thread %d\n",
			kmpt_get_cpu());
		return 1;
	}

	for(int i = 0; i < wctx->papi_num_counters; i++) {
		if(am_event_set_trace_counter(wctx->event_set,
					      wctx->papi_event_mapping[i]+PAPI_COUNTER_BASE,
					      timestamp, wctx->papi_counters[i]))
		{
			fprintf(stderr, "Could not trace counters for thread %d\n",
				kmpt_get_cpu());
			return 1;
		}
	}

	return 0;
}

int kmpt_papi_trace(struct kmpt_worker_context* wctx)
{
	return kmpt_papi_trace_timestamp(wctx, am_timestamp_now(wctx->event_set));
}

#endif
