#ifndef __AFTERMATH_WSSTACK_H__
#define __AFTERMATH_WSSTACK_H__

#include <aftermath/trace/trace.h>
#include <string.h>

enum kmpt_wsstack_entry_type {
	KMPT_WSSTACK_ENTRY_TYPE_FOR = 0,
	KMPT_WSSTACK_ENTRY_TYPE_TASK,
	KMPT_WSSTACK_ENTRY_TYPE_MAX,
	KMPT_WSSTACK_EMPTY
};

/* Entry on the stack representing a loop instance. */
struct kmpt_wsstack_for_entry {
	/* Loop address (e.g., instruction calling dispatch next) */
	void* addr;

	/* ID of the worker's team when the loop was entered */
	uint64_t seq_team_id;

	/* ID of the chunk set the worker was processing */
	am_omp_for_chunk_set_id_t cs_id;
};

/* Entry on the stack representing a task instance. */
struct kmpt_wsstack_task_entry {
	/* Address of the outlined task body */
	void* addr;

	/* ID of the task instance */
	am_omp_task_instance_id_t tiid;
};

static inline size_t kmpt_wsstack_entry_type_size(enum kmpt_wsstack_entry_type t)
{
	switch(t) {
		case KMPT_WSSTACK_ENTRY_TYPE_FOR:
			return sizeof(struct kmpt_wsstack_for_entry);
		case KMPT_WSSTACK_ENTRY_TYPE_TASK:
			return sizeof(struct kmpt_wsstack_task_entry);

		/* Satisfy compiler */
		case KMPT_WSSTACK_ENTRY_TYPE_MAX:
		case KMPT_WSSTACK_EMPTY:
			return 0;
	}

	/* Cannot happen */
	return 0;
}


/* A worksharing stack is a unified stack composed of entries of different types
 * (e.g., kmpt_wsstack_for_entry, kmpt_wsstack_task_entry, etc.). Each stack
 * entry is followed by a kmpt_wsstack_entry_type enum indicating the type of
 * the entry.
 *
 * For example, the memory layout for a stack composed of two task stack entries
 * and a for stack entry is the following:
 *
 *   KMPT_WSSTACK_FOR                  <- (void*)entries + size
 *   struct kmpt_wsstack_for_entry { ... }
 *   KMPT_WSSTACK_TASK
 *   struct kmpt_wsstack_task_entry { ... }
 *   KMPT_WSSTACK_TASK
 *   struct kmpt_wsstack_task_entry { ... }   <- entries
 */
struct kmpt_wsstack {
	/* Base pointer to the raw data of the stack*/
	void* entries;

	/* Number of bytes used */
	size_t used;

	/* Maximum size of the stack in bytes */
	size_t size;

	/* Nesting for each worksharing type */
	int nesting[KMPT_WSSTACK_ENTRY_TYPE_MAX];

	/* Number of entries currently used */
	int top;
};

/* Intialize an empty stack */
static inline int kmpt_wsstack_init(struct kmpt_wsstack* s, int size)
{
	memset(s->nesting, 0, sizeof(s->nesting));

	s->used = 0;
	s->size = size;
	s->top = 0;

	if(!(s->entries = malloc(size)))
		return 1;

	return 0;
}

/* Destroy an worksharing stack */
static inline void kmpt_wsstack_destroy(struct kmpt_wsstack* s)
{
	free(s->entries);
}

/* Reserve a new entry on the stack and returns its address. Returns NULL if the
 * stack does not have enough free memory.
 */
static void* kmpt_wsstack_reserve(struct kmpt_wsstack* s,
				  enum kmpt_wsstack_entry_type type)
{
	void* entry;
	enum kmpt_wsstack_entry_type* type_entry;
	size_t type_size = kmpt_wsstack_entry_type_size(type);
	size_t total_size;

	total_size = type_size + sizeof(enum kmpt_wsstack_entry_type);

	if((s->size - s->used) < total_size)
		return NULL;

	entry = (void*)((long)s->entries + s->used);
	type_entry = (enum kmpt_wsstack_entry_type*)((long)entry + type_size);
	*type_entry = type;

	s->used += total_size;
	s->top++;
	s->nesting[type]++;

	return entry;
}

static inline enum kmpt_wsstack_entry_type
kmpt_wsstack_top_type(struct kmpt_wsstack* s)
{
	enum kmpt_wsstack_entry_type* type_entry;

	if(s->top == 0)
		return KMPT_WSSTACK_EMPTY;

	type_entry = (enum kmpt_wsstack_entry_type*)
		((long)s->entries + s->used -
		 sizeof(enum kmpt_wsstack_entry_type));

	return *type_entry;
}

/* Remove the topmost entry from the stack. Returns 0 on success, otherwise 1
 * (if the stack is empty).
 */
static inline int kmpt_wsstack_pop(struct kmpt_wsstack* s)
{
	enum kmpt_wsstack_entry_type type;
	size_t type_size;
	size_t total_size;

	if((type = kmpt_wsstack_top_type(s)) == KMPT_WSSTACK_EMPTY)
		return 1;

	type_size = kmpt_wsstack_entry_type_size(type);
	total_size = type_size + sizeof(enum kmpt_wsstack_entry_type);

	s->nesting[type]--;
	s->top--;
	s->used -= total_size;

	return 0;
}

/* Return the topmost entry of the stack without popping it. Returns NULL if the
 * stack is empty.
 */
static void* kmpt_wsstack_top(struct kmpt_wsstack* s)
{
	enum kmpt_wsstack_entry_type type;
	size_t type_size;
	size_t total_size;

	if((type = kmpt_wsstack_top_type(s)) == KMPT_WSSTACK_EMPTY)
		return NULL;

	type_size = kmpt_wsstack_entry_type_size(type);
	total_size = type_size + sizeof(enum kmpt_wsstack_entry_type);

	return (void*)((long)s->entries + s->used - total_size);
}

static inline struct kmpt_wsstack_for_entry*
kmpt_wsstack_top_for(struct kmpt_wsstack* s)
{
	return (struct kmpt_wsstack_for_entry*)kmpt_wsstack_top(s);
}

static inline struct kmpt_wsstack_task_entry*
kmpt_wsstack_top_task(struct kmpt_wsstack* s)
{
	return (struct kmpt_wsstack_task_entry*)kmpt_wsstack_top(s);
}

static inline struct kmpt_wsstack_for_entry*
kmpt_wsstack_reserve_for_entry(struct kmpt_wsstack* s)
{
	return (struct kmpt_wsstack_for_entry*)
		kmpt_wsstack_reserve(s, KMPT_WSSTACK_ENTRY_TYPE_FOR);
}

static inline struct kmpt_wsstack_task_entry*
kmpt_wsstack_reserve_task_entry(struct kmpt_wsstack* s)
{
	return (struct kmpt_wsstack_task_entry*)
		kmpt_wsstack_reserve(s, KMPT_WSSTACK_ENTRY_TYPE_TASK);
}

/* Push a new for loop nest onto a stack. Returns 0 on success, 1 otherwise
 * (i.e., if the maximum nesting has been reached). */
static inline int kmpt_wsstack_push_for(struct kmpt_wsstack* s,
					void* addr,
					uint64_t seq_team_id)
{
	struct kmpt_wsstack_for_entry* fe;

	if(!(fe = kmpt_wsstack_reserve_for_entry(s)))
		return 1;

	fe->addr = addr;
	fe->seq_team_id = seq_team_id;

	return 0;
}

/* Push a new task instance onto a stack. Returns 0 on success, 1 otherwise
 * (i.e., if the maximum nesting has been reached). */
static inline int kmpt_wsstack_push_task(struct kmpt_wsstack* s,
					 void* addr)
{
	struct kmpt_wsstack_task_entry* te;

	if(!(te = kmpt_wsstack_reserve_task_entry(s)))
		return 1;

	te->addr = addr;

	return 0;
}

#endif
